#include "headers.h"



/******************************************************************************
  Name:    			128x64 LCD.c package.  Full FONT Version
  Copyright: 		Free to use. not guaranteed to meet any requirements.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Source Code to drive the 128x64, CFAG12864B-YYH-V LCD from
  					http://www.crystalfontz.com/

  Version:			v1.0
*******************************************************************************/



/******************************************************************************
  Name: 			LCD_Initalize(void)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Initalize the LCD
*******************************************************************************/
void LCD_Initalize(void)
{
	LCD_CS1_LOW;										// 
	LCD_CS2_LOW;                                        //
	LCD_E_HIGH;                                         //
	LCD_DI_LOW;                                         //
                                                        //
	LCD_DATA(0x3f);										// Turn Display on
														//
	LCD_DATA(0xc0);										// Start Line at TOP of screen
	                                                    //
	LCD_CS1_HIGH;                                       //
	LCD_CS2_HIGH;                                       //
                                                        //
   	LCD_Cls();                                          //
}
/******************************************************************************
  Name:             E_Pulse_Delay(void)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Delay time for the E line on the LCD
*******************************************************************************/
void E_Pulse_Delay(void)
{
	DWORD x;											// This is Processor specific
                                                        // you need to adjust this for your micro
                                                        //
	for(x=0;x < 5500;x++); 								// ideal speed is 5500, minimum is 4900
}
/******************************************************************************
  Name:             LCD_GetTrueY(BYTE *data, BYTE y)
  Copyright: 		Free to use at will & at own risk.
  Author:           JCP
  Date: 			20/03/05 09:52
  Description:      the LCD Y location is mapped out 0-7 however we need Y from 0-63
*******************************************************************************/
void LCD_GetTrueY(BYTE *data, BYTE y)
{
	BYTE x;												//
                                                        //
	data[0]=1;                                          // start @ 1 then 2,4,8,10,20,40,80
                                                        //
 	for(x=0;x<8;x++)                                    // loop to find which actual Y(0-7) byte
 	{                                                   // for the LCD we are talking about
    	if(y < 8)                                       //
    	{                                               //
            data[0] <<=y;                               // search for the bit position on the true Y
			data[1] = x;                                // location
			return;                                     //
		}                                               //
                                                        //
		y -=8;                                          //
  	}                                                   //
}
/******************************************************************************
  Name:             LCD_PixelPut(BYTE x, BYTE y, BOOL ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      turn on one pixel on the LCD or RAM
*******************************************************************************/
void LCD_PixelPut(BYTE x, BYTE y, BOOL ram_or_lcd)
{

	BYTE value[2];

	if( (x < 128) && (y < 64) )						// protect system ram from user
	{

		LCD_GetTrueY(value,y);
	
		LCD_RAM[x][value[1]] |= value[0];
	 	
	 	if(ram_or_lcd == UPDATE_LCD)
	 	{
			LCD_PutData(x,value[1],LCD_RAM[x][value[1]]);
		}
	}
}
/******************************************************************************
  Name: 			LCD_PixelClr(BYTE x, BYTE y, BOOL ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      turn off one pixel on the LCD or RAM
*******************************************************************************/
void LCD_PixelClr(BYTE x, BYTE y, BOOL ram_or_lcd)
{

 	BYTE value[2];


	if( (x < 128) && (y < 64) )					// protect system ram from user
	{
		LCD_GetTrueY(value,y);
	
		LCD_RAM[x][value[1]] &=~ value[0];
	
	  	if(ram_or_lcd == UPDATE_LCD)
	 	{
			LCD_PutData(x,value[1],LCD_RAM[x][value[1]]);
		}
	}
}
/******************************************************************************
  Name:             LCD_PutData(BYTE x, BYTE y, BYTE data)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Put one byte of data on the LCD. NOTE: Y must be from 0-7, see data sheet
*******************************************************************************/
void LCD_PutData(BYTE x, BYTE y, BYTE data)
{

	if( (x < 128) && (y < 8) )					// protect LCD from user
	{                                           //
		y |=0xB8;								// Set Page Register
	                                            //
		if(x > 0x3f)  							// right side of screen ?
	    {                                       //
	        LCD_CS2_LOW;                        //
	    }                                       //
	                                            //
	    else      								// left side of screen
	    {                                       //
			x |=0x40;							// Set Address Register Bit
	    	LCD_CS1_LOW;                        //
	    }                                       //
                                                //
	                                            //
		LCD_DI_LOW;                             //
	                                            //
		LCD_DATA(y);							// Write Page Register
	                                            //
	    LCD_DATA(x);							// Write Address Register
	                                            //
		LCD_DI_HIGH;							// set display for data mode
	                                            //
		LCD_DATA(data);							// pixel data to put on screen
	                                            //
		LCD_CS1_HIGH;                           // disable CS1
		LCD_CS2_HIGH;                           // disable CS2
                                                //
	}                                           //
}
/******************************************************************************
  Name:         	LCD_Cls(void)
  Copyright: 		Free to use at will & at own risk.
  Author:        	JCP
  Date: 			20/03/05 09:52
  Description:     	clear everything off LCD
*******************************************************************************/
void LCD_Cls(void)
{
	LCD_ClearRam();

	LCD_DisplayRam();
}
/******************************************************************************
  Name:           	LCD_DisplayRam(void)
  Copyright: 		Free to use at will & at own risk.
  Author:         	JCP
  Date: 			20/03/05 09:52
  Description:      To put on LCD everything that is in the LCD RAM buffer
*******************************************************************************/
void LCD_DisplayRam(void)
{
	BYTE col,row;											//
                                                            //
	LCD_CS2_LOW;											// Talk to Controler 2
                                                            //
 	                                                        //
  	for(row=0;row < 8;row++)                                //
	{                                                       //
	    //Talk to both controllers                          //
		LCD_CS1_LOW;										// Talk to Controler 1
                                                            //
		LCD_DI_LOW;											// Instruction Time
                                                            //
    	LCD_DATA(0xB8 + row);								// Set Page counter
                                                            //
    	LCD_DATA(0x40);									    // Set Y address counter
                                                            //
		LCD_DI_HIGH;										// Data Time
                                                            //
                                                            //
	    //Talk to the controller 1                          //
		LCD_CS2_HIGH;                                       //
                                                            //
	    for(col=0;col<0x40;col++)                           //
		{                                                   //
			LCD_DATA(LCD_RAM[col][row]);                    //
        }                                                   //
                                                            //
	    //Talk to the controller 2                          //
		LCD_CS1_HIGH;                                       //
		LCD_CS2_LOW;                                        //
                                                            //
	    for(;col<0x80;col++)                                //
	    {                                                   //
			LCD_DATA(LCD_RAM[col][row]);                    //
        }                                                   //
    }                                                       //
                                                            //
	LCD_CS1_HIGH;                                           //
	LCD_CS2_HIGH;                                           //
}
/******************************************************************************
  Name:           	LCD_ClearRam(void)
  Copyright: 		Free to use at will & at own risk.
  Author:           JCP
  Date: 			20/03/05 09:52
  Description:      To clear the LCD ram buffer
*******************************************************************************/
void LCD_ClearRam(void)
{
	BYTE *point;
	
	point = &LCD_RAM[0][0];
	
 	memset(point,0,1024);

}
/******************************************************************************
  Name:             LCD_FillRam(void)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      To fill the Lcd ram buffer, basically for testing all pixels
*******************************************************************************/
void LCD_FillRam(void)
{
	BYTE *point;
	
	point = &LCD_RAM[0][0];
	
 	memset(point,0xff,1024);
}
/******************************************************************************
  Name: 			LCD_Box(x1, y1, x2, y2, pixel_on_off, ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      To put a BOX on the LCD / RAM, or delete a box off the LCD / RAM
*******************************************************************************/
void LCD_Box(BYTE x1, BYTE y1, BYTE x2, BYTE y2, BOOL pixel_on_off, BOOL ram_or_lcd)
{
	BYTE x,y;

 	if( (x1 < 128) && (x2 <128) && (y1 < 64) && (y2 < 64) )		// protect system from user 
	{
	    x = x1;
	    y = y1;
	
	
	  	while(x <= x2)
	    {
			if(pixel_on_off)
			{
		       	LCD_PixelPut(x,y1,ram_or_lcd);
		       	LCD_PixelPut(x,y2,ram_or_lcd);
			}
			else
			{
		       	LCD_PixelClr(x,y1,ram_or_lcd);
		       	LCD_PixelClr(x,y2,ram_or_lcd);
			}
	
			x++;
	    }
	
	  	while(y <= y2)
	    {
			if(pixel_on_off)
			{
		       	LCD_PixelPut(x1,y,ram_or_lcd);
		       	LCD_PixelPut(x2,y,ram_or_lcd);
			}
			else
			{
		       	LCD_PixelClr(x1,y,ram_or_lcd);
		       	LCD_PixelClr(x2,y,ram_or_lcd);
			}
	
			y++;
	    }
	}
}
/******************************************************************************
  Name: 			LCD_Funct_A(x1, y1, x2, y2, pixel_on_off, ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Used with LCD line function
*******************************************************************************/
void LCD_Funct_A(BYTE x1, BYTE y1, BYTE x2, BYTE y2, BOOL pixel_on_off, BOOL ram_or_lcd)
{
	float a,b;
    BYTE x,y;


    b = (float)((y2 - y1) + 1) / (float)((x2-x1) + 1);

    a = y1;


   	for(x = x1;x <= x2; x++)
    {

    	for(y=(BYTE)a;y <= (BYTE)(a + b); y++)
    	{
			if(y <= y2)
			{
				if(pixel_on_off)
				{
					LCD_PixelPut(x, y, ram_or_lcd);
				}
                else
                {
					LCD_PixelClr(x, y, ram_or_lcd);
                }
			}
        }

	    a+=b;
    }

}
/******************************************************************************
  Name: 			LCD_Funct_B(x1, y1, x2, y2, pixel_on_off, ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      Used with LCD line function
*******************************************************************************/
void LCD_Funct_B(BYTE x1, BYTE y1, BYTE x2, BYTE y2, BOOL pixel_on_off, BOOL ram_or_lcd)
{

	float a,b;
    char x,y;


    b = (float)((y2 - y1) + 1) / (float)((x1-x2) + 1);

    a = y1;


   	for(x = x1;x >= x2; x--)
    {

    	for(y=(BYTE)a;y <= (BYTE)(a + b); y++)
    	{
			if(y <= y2)
			{
				if(pixel_on_off)
				{
					LCD_PixelPut(x, y, ram_or_lcd);
				}
                else
                {
					LCD_PixelClr(x, y, ram_or_lcd);
                }
			}
        }

	    a+=b;
    }

}
/******************************************************************************
  Name: 			LCD_Line(x1, y1, x2, y2, pixel_on_off, ram_or_lcd)
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      To draw or remove a line from the LCD / RAM
*******************************************************************************/
void LCD_Line(BYTE x1, BYTE y1, BYTE x2, BYTE y2, BOOL pixel_on_off,BOOL ram_or_lcd)
{

	if( (x1 < 128) && (x2 < 128) && (y1 < 64) && (y2 < 64) )		// protect system from user 
	{
     	if( (x1 <= x2) && (y1 <= y2) )
     	{
			LCD_Funct_A(x1,y1,x2,y2,pixel_on_off,ram_or_lcd);
		}

     	else if( (x1 > x2)  && (y1 < y2) )
     	{
			LCD_Funct_B(x1,y1,x2,y2,pixel_on_off,ram_or_lcd);
		}

  		else if( (x1 >= x2) && (y1 >= y2) )
  		{
			LCD_Funct_A(x2,y2,x1,y1,pixel_on_off,ram_or_lcd);
  		}

		else if( (x1 < x2) && (y1 > y2) )
		{
			LCD_Funct_B(x2,y2,x1,y1,pixel_on_off,ram_or_lcd);
  		}
	}
}
 /*****************************************************************************
  Name: 			LCD_FixedFont
  Copyright:        Free to use at will & at own risk.
  Author: 			JCP
  Date: 			17/03/05 17:10
  Description:      To Display a fixed size font (6x8 or 8x8) on a 128 x 64 LCD
  Useage:			LCD_Font_Fixed(param 1,param 2,param 3,param 4,param 5)

					param 1: string to be put on the LCD
					param 2: pointer for the font table
     				param 3: 0 = 6x8, 1= 8x8
					param 4: x location in character spaces.  6x8 = 0 - 21 / 8x8 = 0 - 15
					param 5: y location in character spaces.  all fonts Y = 0 - 7

	Example:        sprintf(thestring,"Hello World");
	
         			LCD_FixedFont(thestring,&FONT6x8[0][0],STYLE6x8,0,LINE_1,UPDATE_LCD);

*******************************************************************************/
void LCD_FixedFont(char *the_string, const BYTE *pfont_pointer, BOOL font_size, BYTE location_x, BYTE location_y)
{

	BYTE x,y,z,e,last,pixel_step,loop,x_size,max_x;

	BYTE font[8];

	const BYTE *font_pointer;


 	if(font_size)
 	{
 		x_size = 8;
		max_x = 16;
	} 		
	else
	{
 		x_size = 6;
		max_x = 22;
 	}

   	last = strlen(the_string);								// find the end of the string
                                                            //
	pixel_step = (location_x * x_size);                     // find the result of how many
   	                                                        // pixel spaces we need to increase by
                                                            //
  	for(loop=0;loop < last;loop++)                          //
  	{                                                       //
		font_pointer = pfont_pointer + (8 * (the_string[loop] - 32));                                                            //
															//
	  	z = 0x80;                                           //
                                                            //
		for(x=0;x < x_size;x++ )							// take the x location of font
		{                                                   //
                                                            //
	 		font[x]=0x00;                                   //
                                                            //
	 		e = 0x01;                                       //
                                                            //
			for(y=0;y < 8;y++)								// take the y location of the font
			{                                               //
                                                            //
				if(*font_pointer & z)                       //
				{                                           //
					font[x] |= e;                           //
				}                                           //
                                                            //
				e = e << 1;                                 // rotate up.... 1,2,4,8.....
                                                            //
				font_pointer++;                             //
			}                                               //
                                                            //
			font_pointer -= 8;                              // point back to the top of the font
                                                            //
			z = z >> 1;                                     // rotate down...80,40,20,10....
                                                            //
			LCD_PutData(pixel_step, location_y ,font[x]);	// put data on LCD no RAM
															//		                                                         
			pixel_step++;                                   //
                                                            //
	 	}                                                   //
	}                                                       //
}                                                           //
/******************************************************************************
	Name: 			LCD_FontFreeStyle
  	Copyright:      Free to use at will & at own risk.
  	Author: 		JCP
  	Date: 			17/03/05 17:10
  	Description:    To Display a fixed size font (6x8 or 8x8) in offset Y locations (0 - 63)
  	Useage:			LCD_Font_Fixed(param 1,param 2,param 3,param 4,param 5,param 6)

					param 1: string to be put on the LCD
					param 2: pointer for the font table
     				param 3: 0 = 6x8, 1= 8x8
					param 4: x location in character spaces.  6x8 = 0 - 21 / 8x8 = 0 - 15
					param 5: y location in character spaces.  all fonts = 0 - 63
					param 6: UPDATE_LCD or UPDATE_RAM

	Example:        sprintf(thestring,"Hello World");
	
         			LCD_Free_Style(thestring,&FONT6x8[0][0],STYLE6x8,0,0,UPDATE_LCD);

*******************************************************************************/
void LCD_FontFreeStyle(char *the_string, const BYTE *pfont_pointer, BOOL font_size, BYTE location_x, BYTE location_y, BOOL  ram_or_lcd)
{

	BYTE x,y,z,last,pixel_step,loop,x_size;


	const BYTE *font_pointer;								//
                                                            //
                                                            //
 	if(font_size)                                           // if large font the X spacing must be 8 pixels
 	{                                                       //
 		x_size = 8;                                         //
	} 		                                                //
	else                                                    //
	{                                                       //
 		x_size = 6;                                         // if small font the X spacing must be 6 pixels
 	}                                                       //
                                                            //
                                                            //
  	last = strlen(the_string);								// find the end of the string
                                                            //
	pixel_step = location_x;								//
   	                                                        // pixel spaces we need to increase by
                                                            //
  	for(loop=0;loop < last;loop++)                          //
  	{                                                       //
		font_pointer = pfont_pointer + (8 * (the_string[loop] - 32));                                                            //
															//
	  	z = 0x80;                                           //
                                                            //
		for(x=0;x < x_size;x++ )							// take the x location of font
		{                                                   //
                                                            //
			for(y=0;y < 8;y++)								// take the y location of the font
			{                                               //
                                                            //
				if(*font_pointer & z)						//
				{                                           //
					LCD_PixelPut(pixel_step,location_y + y,ram_or_lcd);  // draw this in pixel mode
				}                                           //
	                                                        //
				font_pointer++;                             //
			}                                               //
                                                            //
			font_pointer -= 8;                              // point back to the top of the font
                                                            //
			z = z >> 1;                                     // rotate down...80,40,20,10....
                                                            //
			pixel_step++;                                   //
	 	}                                                   //
	}                                                       //
}                                                           //
/******************************************************************************
	Name: 			LCD_Logo
  	Copyright:      Free to use at will & at own risk.
  	Author: 		JCP
  	Date: 			17/03/05 17:10
  	Description:    To display LARGE logos on the LCD.
  	Useage:			LCD_Logo(param 1,param 2,param 3,param 4,param 5)

					param 1: pointer for the font table
					param 2: x location on screen, 0 - 127
					param 3: y location on screen, 0 - 63
     				param 4: total 8x8 blocks in the x direction
					param 5: total 8x8 blocks in the y direction

	Info:       	a 128 x 64 LCD gives you a max of 16 (8x8) blocks in the
 					x direction and a max of 8 (8x8) blocks in the y direction.
					When creating a logo draw your logo out in a grid of 128 x 64,
     				then when you are done block it out in a pattern of 8x8 blocks
					and see just how many 8x8 blocks your logo uses. this is how
					you arive with the last 2 parameters of param 4 & 5.

					In this demo the logo LMG took 12 (8x8) blocks in the X direction.
					and 4 (8x8) blocks in the y direction.

	Example:        LCD_Logo(&LMG[0][0],0,20,12,4);
*******************************************************************************/
void LCD_Logo(const BYTE *pfont_pointer, BYTE x_start, BYTE y_start, BYTE x_max, BYTE y_max)
{

	BYTE x,y,z,pixel_step,xloop,yloop,y_track;					//
                                                            	//
	const BYTE *font_pointer;                               	//
                                                            	//
	pixel_step =  x_start;                          			//
                                                                //
	y_track=0;                                      	       	//
	   	                                                    	//
                                                                //
	for(yloop=0;yloop < y_max;yloop++)                          //
	{                                                           //
                                                                //
	  	for(xloop=0;xloop < x_max;xloop++)                    	//
	  	{                                                       //
	                                                            //
			font_pointer = pfont_pointer + (8 * y_track);  	    //
	                                                            //
			y_track++;                                      	//
	                                                            //
		  	z = 0x80;                                           //
		                                                        //
			for(x=0;x < 8;x++ )			                        //
			{                                                   //
	                                                            //
				for(y=0;y < 8;y++)		                        //
				{                                               //
					if(*font_pointer & z)                       //
					{                                           //
						LCD_PixelPut(pixel_step,y_start + y,UPDATE_RAM);
					}                                           //
					else          								// only need this if your going
					{                                           // to scroll the LOGO
						LCD_PixelClr(pixel_step,y_start + y,UPDATE_RAM);
					}
	                                                            //
					font_pointer++;                             //
				}                                               //
		                                                        //
				font_pointer -= 8;                              //
		                                                        //
				z = z >> 1;                                     //
			                                                    //
				pixel_step++;                                   //
		 	}                                                   //
		}                                                       //
	                                                            //
                                                                //
		pixel_step = x_start; 		                         	//
																//
		y_start +=8;                                 			//
                                                                //
	}                                                           //
                                                                //
                                                            	//
    LCD_DisplayRam();											// show it when done writing to ram
																// comment out if you want to do this later
}
/******************************************************************************
  	Name:      		LCD_Circle
  	Copyright: 		Free to use at will & at own risk.
  	Author:    		Found on web @ http://echellon.hybd.net/issues/5/articles/circle/sincos.htm
  	Date:
  	Description:    To Draw a Circle on LCD
  	Useage:			LCD_Circle(param 1,param 2,param 3,param)

					param 1: x location on screen, 0 - 127
					param 2: y location on screen, 0 - 63
					param 3: d diameter of circle

	Example:        LCD_Circle(40,30,30);
*******************************************************************************/
void LCD_Circle(BYTE center_x, BYTE center_y, BYTE r, BOOL ram_or_lcd)
{

	WORD x = 0;
 	WORD y = r;


    while (x <= y)
    {
        // We make use of 8 axes of symmetry in a circle.
        // This way we have fewer points to calculate on its circumference.
        LCD_PixelPut (center_x + (BYTE)x, center_y + (BYTE)y,ram_or_lcd);
        LCD_PixelPut (center_x - (BYTE)x, center_y + (BYTE)y,ram_or_lcd);
        LCD_PixelPut (center_x + (BYTE)x, center_y - (BYTE)y,ram_or_lcd);
        LCD_PixelPut (center_x - (BYTE)x, center_y - (BYTE)y,ram_or_lcd);
        LCD_PixelPut (center_x + (BYTE)y, center_y + (BYTE)x,ram_or_lcd);
        LCD_PixelPut (center_x - (BYTE)y, center_y + (BYTE)x,ram_or_lcd);
        LCD_PixelPut (center_x + (BYTE)y, center_y - (BYTE)x,ram_or_lcd);
        LCD_PixelPut (center_x - (BYTE)y, center_y - (BYTE)x,ram_or_lcd);

        // This is the most important part of the function.
        // We go to the right in all cases (x++).
        // We need to decide whether to go down (y--).
        // This depends on which point is
        // closest to the path of the circle.
        // Good old Pythagoras will tell us what to do.
        x++;

        if( abs(x*x + y*y - r*r) > abs(x*x + (y-1)*(y-1) - r*r) )
        {
	        y--;
		}
    }
}
/******************************************************************************
  Name:      		LCD_Test
  Copyright: 		Free to use at will & at own risk.
  Author: 			JCP
  Date: 			20/03/05 09:52
  Description:      To draw pixels on screen In a kinda cool way
*******************************************************************************/
void LCD_Test(void)
{

	int x,y;


   	for(x=0;x<= 127;x++)
   	{
   		LCD_Line(x,0,63,32, ON, UPDATE_LCD);
    }


  	for(y=0;y< 64;y++)
   	{
   		LCD_Line(127,y,63,32, ON, UPDATE_LCD);
    }

   	for(x=127;x >= 0;x--)
   	{
   		LCD_Line(x,63,63,32, ON, UPDATE_LCD);
    }


  	for(y=63;y >=0;y--)
   	{
  		LCD_Line(0,y,63,32, ON, UPDATE_LCD);
    }




   	for(x=0;x<= 127;x++)
   	{
   		LCD_Line(x,0,63,32, OFF, UPDATE_LCD);
    }


  	for(y=0;y< 64;y++)
   	{
   		LCD_Line(127,y,63,32, OFF, UPDATE_LCD);
    }

   	for(x=127;x >= 0;x--)
  	{
   		LCD_Line(x,63,63,32, OFF, UPDATE_LCD);
    }


  	for(y=63;y >=0;y--)
   	{
   		LCD_Line(0,y,63,32, OFF, UPDATE_LCD);
    }
}
